"""user.py contains basic information about the steno3d user"""

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals

from .traits import HasSteno3DTraits, Int, String


class User(HasSteno3DTraits):
    """Class representing a user instance"""
    _model_api_location = "user"

    email = String(
        help='Email',
        default_value=None,
        allow_none=True,
        read_only=True
    )
    name = String(
        help='Name',
        default_value=None,
        allow_none=True,
        read_only=True
    )
    url = String(
        help='URL',
        default_value=None,
        allow_none=True,
        read_only=True
    )
    affiliation = String(
        help='Affiliation',
        default_value=None,
        allow_none=True,
        read_only=True
    )
    location = String(
        help='Location',
        default_value=None,
        allow_none=True,
        read_only=True
    )
    username = String(
        help='Username',
        default_value=None,
        allow_none=True,
        read_only=True
    )

    devel_key = String(
        help='Developer API Key',
        default_value=None,
        allow_none=True,
        read_only=True
    )

    file_size_limit = Int(
        help='Inidividual file limit',
        default_value=5000000,
        read_only=True
    )
    project_size_limit = Int(
        help='Project size limit',
        default_value=25000000,
        read_only=True
    )
    project_resource_limit = Int(
        help='Maximum resources in a project',
        default_value=25,
        read_only=True
    )

    def login_with_json(self, login_json):
        self.set_trait('username', login_json['uid'])
        self.set_trait('email', login_json['email'])
        self.set_trait('name', login_json['name'])
        self.set_trait('url', login_json['url'])
        self.set_trait('affiliation', login_json['affiliation'])
        self.set_trait('location', login_json['location'])

    def set_key(self, devel_key):
        self.set_trait('devel_key', devel_key)

    def logout(self):
        self.set_trait('username', None)
        self.set_trait('email', None)
        self.set_trait('name', None)
        self.set_trait('url', None)
        self.set_trait('affiliation', None)
        self.set_trait('location', None)
        self.set_trait('devel_key', None)

    @property
    def logged_in(self):
        return self.username is not None



"""base.py sets up BaseExample class that allows examples to be accessed
as properties at the class level
"""

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals

from os import makedirs, mkdir
from os.path import exists, expanduser, isdir, realpath, sep

from requests import get
from six import string_types
from zipfile import ZipFile


class exampleproperty(object):
    """wrapper that sets class method as property"""

    def __init__(self, func):
        self.func = classmethod(func)

    def __get__(self, cls, owner):
        return self.func.__get__(None, owner)()


class BaseExample(object):
    """basic class that all examples inherit from"""

    def __init__(self, *args, **kwargs):
        raise TypeError('Examples cannot be instantiated. Please access '
                        'the properties directly.')

    @exampleproperty
    def example_name(self):
        return self.__name__

    @exampleproperty
    def data_directory(self):
        """path to directory containing all assets"""
        return sep.join([expanduser('~'), '.steno3d_client', 'assets'])

    @exampleproperty
    def sub_directory(self):
        return self.example_name.lower()

    @exampleproperty
    def filenames(self):
        return []

    @exampleproperty
    def data_url(self):
        return 'https://storage.googleapis.com/steno3d-examples'

    @classmethod
    def fetch_data(cls, directory=None, download_if_missing=True,
                   filename=None, verbose=True):
        """method fetch_data

        This method checks for example data locally and downloads and
        extracts the archive if necessary.

        Inputs:
            directory:           Local folder to save the archive and
                                 extracted data.
                                 (default: ~/.steno3d_client/assets/)
            download_if_missing: Download the data archive if it is not
                                 found locally. (default: True)
            filename:            Specific data file to fetch. If none,
                                 fetch all example data. (default: None)
            verbose:             Print download/extraction status.
                                 (default: True)
        """

        if filename is not None and not isinstance(filename, string_types):
            raise ValueError('filename: must be the name of one file')
        if cls.filenames == []:
            raise ValueError('Example does not require any files')
        if directory is not None and not isinstance(directory, string_types):
            raise ValueError('directory: must be the name of a directory')
        if directory is not None:
            directory = realpath(expanduser(directory))
            if not isdir(directory):
                raise ValueError(
                    '{}: directory does not exist'.format(directory)
                )
            cls._data_directory = directory
        if verbose:
            print('Fetching data...')
        destination = sep.join([cls.data_directory, cls.sub_directory])
        archive = sep.join([destination, cls.sub_directory + '.zip'])
        if download_if_missing and not exists(cls.data_directory):
            makedirs(cls.data_directory)
        if download_if_missing and not exists(destination):
            mkdir(destination)
        if verbose:
            print('Destination: ' + destination)
        filenames = cls.filenames if filename is None else [filename]
        for fname in filenames:
            if verbose:
                print('    Fetching: ' + fname)
            destination_file = sep.join([destination, fname])
            if exists(destination_file):
                if verbose:
                    print('... Local copy found')
                if filename is not None:
                    return destination_file
                continue
            if not exists(archive) and download_if_missing:
                if verbose:
                    print('        Downloading archive...')
                url = '/'.join([cls.data_url, cls.sub_directory + '.zip'])
                resp = get(url, stream=True)
                if resp.status_code != 200:
                    raise IOError('Error downloading {exclass} data: '
                                  '{archfile}'.format(
                                    exclass=cls.example_name,
                                    archfile=archive
                                  ))
                with open(archive, 'wb') as arch:
                    for chunk in resp:
                        arch.write(chunk)
                if verbose:
                    print('        Archive downloaded successfully!')
            if exists(archive):
                if verbose:
                    print('        Local archive found: extracting...')
                try:
                    zf = ZipFile(archive)
                    zf.extract(fname, destination)
                    if verbose:
                        print('... File extracted successfully!')
                    if filename is not None:
                        return destination_file
                except KeyError:
                    raise IOError(
                        """Required file(s) not found in archive:
                        Archive file may be out of date. Please delete archive
                        and {exclass}.fetch_data() again: {archfile}""".format(
                            exclass=cls.example_name,
                            archfile=archive
                        )
                    )
                continue
            raise IOError(
                """Required file(s) not found:
                Please call {exclass}.fetch_data() to download and
                save to a default folder in your home directory or
                {exclass}.fetch_data(directory=your_local_directory)
                to set an alternative local directory.""".format(
                    exclass=cls.example_name
                )
            )



"""wolfpass.py is a project example using resources from Wolf Pass"""

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals

from numpy import load as npload

from .base import BaseExample, exampleproperty
from ..data import DataArray
from ..line import Line, Mesh1D
from ..point import Mesh0D, Point
from ..project import Project
from ..surface import Mesh2D, Surface
from ..texture import Texture2DImage
from ..volume import Mesh3DGrid, Volume


class Wolfpass(BaseExample):
    """Wolfpass example

    This module provides example data for a geology exploration project.
    """

    @exampleproperty
    def filenames(self):
        return [
            'AG_gpt.line.npy',
            'AS_ppm.line.npy',
            'AU_gpt.line.npy',
            'CU_pct.line.npy',
            'CU_pct.vol.npy',
            'CU_pct_0.75_1.0_t.cusurf.npy',
            'CU_pct_0.75_1.0_v.cusurf.npy',
            'CU_pct_1.0_1.25_t.cusurf.npy',
            'CU_pct_1.0_1.25_v.cusurf.npy',
            'CU_pct_1.25_1.5_t.cusurf.npy',
            'CU_pct_1.25_1.5_v.cusurf.npy',
            'CU_pct_gt_1.5_t.cusurf.npy',
            'CU_pct_gt_1.5_v.cusurf.npy',
            'CU_pct_lt_0.75_t.cusurf.npy',
            'CU_pct_lt_0.75_v.cusurf.npy',
            'Density.line.npy',
            'MO_ppm.line.npy',
            'Recov.line.npy',
            'S_pct.line.npy',
            'basement_t.lithsurf.npy',
            'basement_v.lithsurf.npy',
            'boreholes_s.line.npy',
            'boreholes_v.line.npy',
            'dacite_data.line.npy',
            'dacite_t.lithsurf.npy',
            'dacite_v.lithsurf.npy',
            'diorite_early_t.lithsurf.npy',
            'diorite_early_v.lithsurf.npy',
            'diorite_late_t.lithsurf.npy',
            'diorite_late_v.lithsurf.npy',
            'dist_to_borehole.lithsurf.npy',
            'dist_to_borehole.vol.npy',
            'drill_loc_v.point.npy',
            'elevation.toposurf.npy',
            'lithology.xsurf.npy',
            'maxdepth.point.npy',
            'ovb_t.lithsurf.npy',
            'ovb_v.lithsurf.npy',
            'section_number.xsurf.npy',
            'topo_t.toposurf.npy',
            'topo_v.toposurf.npy',
            'topography.png',
            'trench.point.npy',
            'vol_h1.vol.npy',
            'vol_h2.vol.npy',
            'vol_h3.vol.npy',
            'vol_x0.vol.npy',
            'xsect_t.xsurf.npy',
            'xsect_v.xsurf.npy',
        ]

    @exampleproperty
    def drill_vertices(self):
        """drill point vertices"""
        return npload(Wolfpass.fetch_data(filename='drill_loc_v.point.npy',
                                          download_if_missing=False,
                                          verbose=False))

    @exampleproperty
    def drill_data(self):
        """dictionry of drill point data"""
        if getattr(self, '_drill_data', None) is None:
            self._drill_data = dict()
            for npyfile in self.filenames:
                if not npyfile.endswith('.point.npy'):
                    continue
                if npyfile.endswith('_v.point.npy'):
                    continue
                self._drill_data[npyfile.split('.')[0]] = npload(
                    Wolfpass.fetch_data(filename=npyfile,
                                        download_if_missing=False,
                                        verbose=False)
                )
        return self._drill_data

    @exampleproperty
    def borehole_vertices(self):
        """borehole line vertices"""
        return npload(Wolfpass.fetch_data(filename='boreholes_v.line.npy',
                                          download_if_missing=False,
                                          verbose=False))

    @exampleproperty
    def borehole_segments(self):
        """borehole segment vertex indices"""
        return npload(Wolfpass.fetch_data(filename='boreholes_s.line.npy',
                                          download_if_missing=False,
                                          verbose=False))

    @exampleproperty
    def borehole_data(self):
        """dictionary of borehole data"""
        if getattr(self, '_borehole_data', None) is None:
            self._borehole_data = dict()
            for npyfile in self.filenames:
                if not npyfile.endswith('.line.npy'):
                    continue
                if (npyfile.endswith('_v.line.npy') or
                        npyfile.endswith('_s.line.npy')):
                    continue
                self._borehole_data[npyfile.split('.')[0]] = npload(
                    Wolfpass.fetch_data(filename=npyfile,
                                        download_if_missing=False,
                                        verbose=False)
                )
        return self._borehole_data

    @exampleproperty
    def cu_names(self):
        """list of names for the different cu pct surfaces"""
        return [fname[:-13] for fname in self.filenames
                if fname.endswith('_v.cusurf.npy')]

    @exampleproperty
    def cu_vertices(self):
        """list of cu pct surface vertices"""
        return [npload(Wolfpass.fetch_data(filename=prefix + '_v.cusurf.npy',
                                           download_if_missing=False,
                                           verbose=False))
                for prefix in self.cu_names]

    @exampleproperty
    def cu_triangles(self):
        """list of cu pct surface triangles"""
        return [npload(Wolfpass.fetch_data(filename=prefix + '_t.cusurf.npy',
                                           download_if_missing=False,
                                           verbose=False))
                for prefix in self.cu_names]

    @exampleproperty
    def lith_names(self):
        """list of names for the different lithology surfaces"""
        return [fname[:-15] for fname in self.filenames
                if fname.endswith('_v.lithsurf.npy')]

    @exampleproperty
    def lith_vertices(self):
        """list of lithology surface vertices"""
        return [npload(Wolfpass.fetch_data(filename=prefix + '_v.lithsurf.npy',
                                           download_if_missing=False,
                                           verbose=False))
                for prefix in self.lith_names]

    @exampleproperty
    def lith_triangles(self):
        """list of lithology surface triangles"""
        return [npload(Wolfpass.fetch_data(filename=prefix + '_t.lithsurf.npy',
                                           download_if_missing=False,
                                           verbose=False))
                for prefix in self.lith_names]

    @exampleproperty
    def lith_diorite_early_data(self):
        """data for early diorite surface"""
        return npload(Wolfpass.fetch_data(
            filename='dist_to_borehole.lithsurf.npy',
            download_if_missing=False,
            verbose=False
        ))

    @exampleproperty
    def topo_vertices(self):
        """topography vertices"""
        return npload(Wolfpass.fetch_data(filename='topo_v.toposurf.npy',
                                          download_if_missing=False,
                                          verbose=False))

    @exampleproperty
    def topo_triangles(self):
        """topography triangles"""
        return npload(Wolfpass.fetch_data(filename='topo_t.toposurf.npy',
                                          download_if_missing=False,
                                          verbose=False))

    @exampleproperty
    def topo_image(self):
        """surface image PNG"""
        return Wolfpass.fetch_data(filename='topography.png',
                                   download_if_missing=False,
                                   verbose=False)

    @exampleproperty
    def topo_image_orientation(self):
        """surface image O, U, and V"""
        return dict(
            O=[443200., 491750, 0],
            U=[4425., 0, 0],
            V=[0., 3690, 0]
        )

    @exampleproperty
    def topo_data(self):
        """elevation data"""
        return npload(Wolfpass.fetch_data(filename='elevation.toposurf.npy',
                                          download_if_missing=False,
                                          verbose=False))

    @exampleproperty
    def xsect_vertices(self):
        """cross section vertices"""
        return npload(Wolfpass.fetch_data(filename='xsect_v.xsurf.npy',
                                          download_if_missing=False,
                                          verbose=False))

    @exampleproperty
    def xsect_triangles(self):
        """cross section triangles"""
        return npload(Wolfpass.fetch_data(filename='xsect_t.xsurf.npy',
                                          download_if_missing=False,
                                          verbose=False))

    @exampleproperty
    def xsect_data(self):
        """dictionary of cross section data"""
        if getattr(self, '_xsect_data', None) is None:
            self._xsect_data = dict()
            for npyfile in self.filenames:
                if not npyfile.endswith('.xsurf.npy'):
                    continue
                if (npyfile.endswith('_v.xsurf.npy') or
                        npyfile.endswith('_t.xsurf.npy')):
                    continue
                self._xsect_data[npyfile.split('.')[0]] = npload(
                    Wolfpass.fetch_data(filename=npyfile,
                                        download_if_missing=False,
                                        verbose=False)
                )
        return self._xsect_data

    @exampleproperty
    def lith_tensor(self):
        """h1, h2, h3 dictionary for lith volume"""
        return dict(
            h1=npload(Wolfpass.fetch_data(filename='vol_h1.vol.npy',
                                          download_if_missing=False,
                                          verbose=False)),
            h2=npload(Wolfpass.fetch_data(filename='vol_h2.vol.npy',
                                          download_if_missing=False,
                                          verbose=False)),
            h3=npload(Wolfpass.fetch_data(filename='vol_h3.vol.npy',
                                          download_if_missing=False,
                                          verbose=False))
        )

    @exampleproperty
    def lith_origin(self):
        """x0 for lith volume"""
        return npload(Wolfpass.fetch_data(filename='vol_x0.vol.npy',
                                          download_if_missing=False,
                                          verbose=False))

    @exampleproperty
    def lith_data(self):
        """dictionary of data for lith volume"""
        if getattr(self, '_lith_data', None) is None:
            self._lith_data = dict()
            for npyfile in self.filenames:
                if not npyfile.endswith('.vol.npy'):
                    continue
                if npyfile.startswith('vol_'):
                    continue
                self._lith_data[npyfile.split('.')[0]] = npload(
                    Wolfpass.fetch_data(filename=npyfile,
                                        download_if_missing=False,
                                        verbose=False)
                ).flatten()
        return self._lith_data

    @classmethod
    def get_project(self):
        """Return a project with all the Wolf Pass data"""
        proj = Project(
            title='Wolf Pass'
        )
        self._add_points(proj)
        self._add_lines(proj)
        self._add_cu_surf(proj)
        self._add_lith_surf(proj)
        self._add_topo(proj)
        self._add_xsect(proj)
        self._add_lith_vol(proj)
        return proj

    @classmethod
    def get_project_topo(self):
        """Return a project with Wolf Pass topography data"""
        proj = Project(
            title='Topography',
            description='Topography, surface imagery, and drill locations'
        )
        self._add_points(proj)
        self._add_topo(proj)
        return proj

    @classmethod
    def get_project_dacite(self):
        """Return a project with Wolf Pass dacite data"""
        proj = Project(
            title='Wolf Pass',
            description='Boreholes and dacite formation'
        )
        self._add_lines(proj)
        self._add_lith_surf(proj, ['ovb', 'dacite'])
        return proj

    @classmethod
    def _add_points(self, proj):
        Point(
            project=proj,
            mesh=Mesh0D(
                vertices=self.drill_vertices
            ),
            data=[
                dict(
                    location='N',
                    data=DataArray(
                        title=k,
                        array=self.drill_data[k]
                    )
                ) for k in self.drill_data
            ],
            title='Borehole Drill Locations'
        )

    @classmethod
    def _add_lines(self, proj):
        Line(
            project=proj,
            mesh=Mesh1D(
                vertices=self.borehole_vertices,
                segments=self.borehole_segments
            ),
            data=[
                dict(
                    location='CC',
                    data=DataArray(
                        title=k,
                        array=self.borehole_data[k]
                    )
                ) for k in self.borehole_data
            ],
            title='Boreholes'
        )

    @classmethod
    def _add_cu_surf(self, proj):
        for i, prefix in enumerate(self.cu_names):
            Surface(
                project=proj,
                mesh=Mesh2D(
                    vertices=self.cu_vertices[i],
                    triangles=self.cu_triangles[i]
                ),
                title=prefix
            )

    @classmethod
    def _add_lith_surf(self, proj, include=None):
        for i, prefix in enumerate(self.lith_names):
            if include is not None and prefix not in include:
                continue
            if prefix == 'diorite_early':
                lith_data = [dict(
                                location='N',
                                data=DataArray(
                                    title='Distance to Borehole',
                                    array=self.lith_diorite_early_data
                                )
                            )]
            else:
                lith_data = []
            Surface(
                project=proj,
                mesh=Mesh2D(
                    vertices=self.lith_vertices[i],
                    triangles=self.lith_triangles[i]
                ),
                data=lith_data,
                title=prefix
            )

    @classmethod
    def _add_topo(self, proj):
        Surface(
            project=proj,
            mesh=Mesh2D(
                vertices=self.topo_vertices,
                triangles=self.topo_triangles
            ),
            data=dict(
                location='N',
                data=DataArray(
                    title='Elevation',
                    array=self.topo_data
                )
            ),
            textures=Texture2DImage(
                image=self.topo_image,
                **self.topo_image_orientation
            ),
            title='Topography Surface'
        )

    @classmethod
    def _add_xsect(self, proj):
        Surface(
            project=proj,
            mesh=Mesh2D(
                vertices=self.xsect_vertices,
                triangles=self.xsect_triangles
            ),
            data=[
                dict(
                    location='CC',
                    data=DataArray(
                        title=k,
                        array=self.xsect_data[k]
                    )
                ) for k in self.xsect_data
            ],
            title='Cross-Sections'
        )

    @classmethod
    def _add_lith_vol(self, proj):
        Volume(
            project=proj,
            mesh=Mesh3DGrid(
                x0=self.lith_origin,
                **self.lith_tensor
            ),
            data=[
                dict(
                    location='CC',
                    data=DataArray(
                        title=k,
                        array=self.lith_data[k]
                    )
                ) for k in self.lith_data
            ],
            title='Lithology Volume'
        )



from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals

from six import integer_types

from .client import Comms, needs_login
from .project import Project


MINE = 'api/project/steno3ds/mine'


def _query(url, queue=10):
    cursor = ''
    more = True
    print('Fetching projects from the database ...')
    while more:
        resp = Comms.get('{url}?brief=True&num={n}&cursor={c}'.format(
            url=url, n=queue, c=cursor
        ))
        rjson = resp.json()
        cursor = rjson['cursor']
        more = rjson['more']
        for proj in rjson['data']:
            yield proj
        print('Fetching more projects from the database ...')


def _short_json(proj_json):
    return {'uid': proj_json['uid'],
            'title': proj_json['title'],
            'description': proj_json['description'],
            'created': proj_json['date']}


@needs_login
def my_projects(n=None, queue=100):
    if n is None:
        print('Querying all your projects ...')
        return [_short_json(p) for p in _query(MINE, queue)]
    if not isinstance(n, integer_types):
        raise ValueError('{}: n must be int'.format(n))
    print('Querying your most recent {} project(s) ...'.format(n))
    projs = []
    projit = _query(MINE, min(n, queue))
    for _ in range(n):
        try:
            projs += [_short_json(next(projit))]
        except StopIteration:
            print('{n}: n > total number of projects, {p} returned'.format(
                n=n, p=len(projs)
            ))
            break
    print('...Complete!')
    return projs


@needs_login
def project_by_uid(uid, copy=None):
    return Project._build(uid, copy)


@needs_login
def last_project(copy=None):
    try:
        return project_by_uid(next(_query(MINE, 1))['uid'], copy)
    except StopIteration:
        print('No projects available!')



"""data.py contains resource data structures"""

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals

from six import string_types

from numpy import ndarray
from traitlets import observe, validate

from .base import BaseData
from .traits import Array, String


class DataArray(BaseData):
    """Data array with unique values at every point in the mesh"""
    _resource_class = 'array'
    array = Array(
        help='Data, unique values at every point in the mesh',
        shape=('*',),
        dtype=(float, int)
    )

    order = String(
        help='Data array order, for data on grid meshes',
        choices={
            'c': ('C-STYLE', 'NUMPY', 'ROW-MAJOR', 'ROW'),
            'f': ('FORTRAN', 'MATLAB', 'COLUMN-MAJOR', 'COLUMN', 'COL')
        },
        default_value='c',
        lowercase=True
    )

    def __init__(self, array=None, **kwargs):
        super(DataArray, self).__init__(**kwargs)
        if array is not None:
            self.array = array

    def _nbytes(self, arr=None):
        if arr is None or (isinstance(arr, string_types) and arr == 'array'):
            arr = self.array
        if isinstance(arr, ndarray):
            return arr.astype('f4').nbytes
        raise ValueError('DataArray cannot calculate the number of '
                         'bytes of {}'.format(arr))

    @observe('array')
    def _reject_large_files(self, change):
        try:
            self._validate_file_size(change['name'], change['new'])
        except ValueError as err:
            setattr(change['owner'], change['name'], change['old'])
            raise err

    @validate('array')
    def _validate_array(self, proposal):
        proposal['owner']._validate_file_size('array', proposal['value'])
        return proposal['value']

    def _get_dirty_data(self, force=False):
        datadict = super(DataArray, self)._get_dirty_data(force)
        dirty = self._dirty_traits
        if 'order' in dirty or force:
            datadict['order'] = self.order
        return datadict

    def _get_dirty_files(self, force=False):
        files = super(DataArray, self)._get_dirty_files(force)
        dirty = self._dirty_traits
        if 'array' in dirty or force:
            files['array'] = self.traits()['array'].serialize(self.array)
        return files

    @classmethod
    def _build_from_json(cls, json, **kwargs):
        data = DataArray(
            title=kwargs['title'],
            description=kwargs['description'],
            order=json['order'],
            array=Array.download(
                url=json['array'],
                shape=json['arraySize']//4,
                dtype=json['arrayType']
            )
        )
        return data

    @classmethod
    def _build_from_omf(cls, omf_data):
        assert omf_data.__class__.__name__ in ('ScalarData', 'MappedData')
        data = dict(
            location='N' if omf_data.location == 'vertices' else 'CC',
            data=DataArray(
                title=omf_data.name,
                description=omf_data.description,
                array=omf_data.array.array
            )
        )
        return data

__all__ = ['DataArray']



"""line.py contains the Line composite resource for steno3d"""

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals

from numpy import max as npmax
from numpy import min as npmin
from numpy import ndarray
from six import string_types
from traitlets import observe, validate

from .base import BaseMesh
from .base import CompositeResource
from .data import DataArray
from .options import ColorOptions
from .options import Options
from .traits import Array, HasSteno3DTraits, KeywordInstance, Repeated, String


class _Mesh1DOptions(Options):
    view_type = String(
        help='Display 1D lines or tubes/boreholes/extruded lines',
        choices={
            'line': ('lines', 'thin', '1d'),
            'tube': ('tubes', 'extruded line', 'extruded lines',
                     'borehole', 'boreholes')
        },
        default_value='line',
        lowercase=True,
        allow_none=True
    )


class _LineOptions(ColorOptions):
    pass


class Mesh1D(BaseMesh):
    """Contains spatial information of a 1D line set"""
    vertices = Array(
        help='Mesh vertices',
        shape=('*', 3),
        dtype=float
    )
    segments = Array(
        help='Segment endpoint indices',
        shape=('*', 2),
        dtype=int
    )
    opts = KeywordInstance(
        help='Options',
        klass=_Mesh1DOptions,
        allow_none=True
    )

    @property
    def nN(self):
        """ get number of nodes """
        return len(self.vertices)

    @property
    def nC(self):
        """ get number of cells """
        return len(self.segments)

    def _nbytes(self, arr=None):
        if arr is None:
            return self._nbytes('segments') + self._nbytes('vertices')
        if isinstance(arr, string_types) and arr in ('segments', 'vertices'):
            arr = getattr(self, arr)
        if isinstance(arr, ndarray):
            return arr.astype('f4').nbytes
        raise ValueError('Mesh1D cannot calculate the number of '
                         'bytes of {}'.format(arr))

    @observe('segments', 'vertices')
    def _reject_large_files(self, change):
        try:
            self._validate_file_size(change['name'], change['name'])
        except ValueError as err:
            setattr(change['owner'], change['name'], change['old'])
            raise err

    @validate('segments')
    def _validate_seg(self, proposal):
        if npmin(proposal['value']) < 0:
            raise ValueError('Segments may only have positive integers')
        if npmax(proposal['value']) >= len(proposal['owner'].vertices):
            raise ValueError('Segments expects more vertices than provided')
        proposal['owner']._validate_file_size('segments', proposal['value'])
        return proposal['value']

    @validate('vertices')
    def _validate_vert(self, proposal):
        if npmax(proposal['owner'].segments) >= len(proposal['value']):
            raise ValueError('Segments expects more vertices than provided')
        proposal['owner']._validate_file_size('vertices', proposal['value'])
        return proposal['value']

    def _get_dirty_files(self, force=False):
        files = super(Mesh1D, self)._get_dirty_files(force)
        dirty = self._dirty_traits
        if 'vertices' in dirty or force:
            files['vertices'] = \
                self.traits()['vertices'].serialize(self.vertices)
        if 'segments' in dirty or force:
            files['segments'] = \
                self.traits()['segments'].serialize(self.segments)
        return files

    @classmethod
    def _build_from_json(cls, json, **kwargs):
        mesh = Mesh1D(
            title=kwargs['title'],
            description=kwargs['description'],
            vertices=Array.download(
                url=json['vertices'],
                shape=(json['verticesSize']//12, 3),
                dtype=json['verticesType']
            ),
            segments=Array.download(
                url=json['segments'],
                shape=(json['segmentsSize']//8, 2),
                dtype=json['segmentsType']
            ),
            opts=json['meta']
        )
        return mesh

    @classmethod
    def _build_from_omf(cls, omf_geom, omf_project):
        mesh = Mesh1D(
            vertices=(omf_geom.vertices.array +
                      omf_geom.origin +
                      omf_project.origin),
            segments=omf_geom.segments.array
        )
        return mesh


class _LineBinder(HasSteno3DTraits):
    """Contains the data on a 1D line set with location information"""
    location = String(
        help='Location of the data on mesh',
        choices={
            'CC': ('LINE', 'FACE', 'CELLCENTER', 'EDGE', 'SEGMENT'),
            'N': ('VERTEX', 'NODE', 'ENDPOINT')
        }
    )
    data = KeywordInstance(
        help='Data',
        klass=DataArray
    )


class Line(CompositeResource):
    """Contains all the information about a 1D line set"""
    mesh = KeywordInstance(
        help='Mesh',
        klass='Mesh1D'
    )
    data = Repeated(
        help='Data',
        trait=KeywordInstance(klass=_LineBinder),
        allow_none=True
    )
    opts = KeywordInstance(
        help='Options',
        klass=_LineOptions,
        allow_none=True
    )

    def _nbytes(self):
        return self.mesh._nbytes() + sum(d.data._nbytes() for d in self.data)

    @validate('data')
    def _validate_data(self, proposal):
        """Check if resource is built correctly"""
        for ii, dat in enumerate(proposal['value']):
            assert dat.location in ('N', 'CC')
            valid_length = (
                proposal['owner'].mesh.nC if dat.location == 'CC'
                else proposal['owner'].mesh.nN
            )
            if len(dat.data.array) != valid_length:
                raise ValueError(
                    'line.data[{index}] length {datalen} does not match '
                    '{loc} length {meshlen}'.format(
                        index=ii,
                        datalen=len(dat.data.array),
                        loc=dat.location,
                        meshlen=valid_length
                    )
                )
        return proposal['value']


__all__ = ['Line', 'Mesh1D']



from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals

import json

import numpy as np

from .base import BaseExample, exampleproperty
from ..data import DataArray
from ..project import Project
from ..surface import Mesh2D, Mesh2DGrid, Surface
from ..texture import Texture2DImage


class Topography(BaseExample):
    """Topography example

    This module provides example data for a topography surface.
    """

    @exampleproperty
    def filenames(self):
        return ['topo_grid.json', 'topo_tris.json', 'topo.png']

    @exampleproperty
    def gridfile(self):
        return Topography.fetch_data(filename='topo_grid.json',
                                     download_if_missing=False,
                                     verbose=False)

    @exampleproperty
    def trifile(self):
        return Topography.fetch_data(filename='topo_tris.json',
                                     download_if_missing=False,
                                     verbose=False)

    @exampleproperty
    def griddata(self):
        """topography data from json"""
        with open(Topography.gridfile, 'r') as f:
            topo = json.load(f)
        return topo

    @exampleproperty
    def tridata(self):
        """topography data from json"""
        with open(Topography.trifile, 'r') as f:
            topo = json.load(f)
        return topo

    @exampleproperty
    def topo_image(self):
        """surface image"""
        return Topography.fetch_data(filename='topo.png',
                                     download_if_missing=False,
                                     verbose=False)

    @exampleproperty
    def topo_image_orientation(self):
        """surface image O, U, and V"""
        return dict(
            O=[443200., 491750, 0],
            U=[4425., 0, 0],
            V=[0., 3690, 0]
        )

    @classmethod
    def get_project(self):
        """return topography project"""
        elev_n = np.array(Topography.tridata['vertices'])[:, 2]
        elev_cc = np.mean(
            elev_n[np.array(Topography.tridata['triangles'])],
            axis=1
        )
        proj = Project(
            title='Topography'
        )
        Surface(
            project=proj,
            mesh=Mesh2D(
                vertices=Topography.tridata['vertices'],
                triangles=Topography.tridata['triangles']
            ),
            data=[
                dict(
                    location='N',
                    data=DataArray(
                        array=elev_n,
                        title='Elevation, vertices'
                    )
                ),
                dict(
                    location='CC',
                    data=DataArray(
                        array=elev_cc,
                        title='Elevation, faces'
                    )
                )
            ],
            textures=[
                Texture2DImage(
                    image=Topography.topo_image,
                    **Topography.topo_image_orientation
                )
            ],
            title='Topography Tri-Surface',
            description=('This surface has face and vertex elevation '
                         'data as well as surface imagery')
        )
        Surface(
            project=proj,
            mesh=Mesh2DGrid(
                h1=np.diff(Topography.griddata['x']),
                h2=np.diff(Topography.griddata['y']),
                Z=Topography.griddata['z'],
                O=[Topography.griddata['x'][0], Topography.griddata['y'][0], 0]
            ),
            data=[
                dict(
                    location='N',
                    data=DataArray(
                        array=Topography.griddata['z'],
                        title='Elevation, vertices'
                    )
                ),
            ],
            textures=[
                Texture2DImage(
                    image=Topography.topo_image,
                    **Topography.topo_image_orientation
                )
            ],
            title='Topography Grid-Surface',
            description=('This surface has vertex elevation '
                         'data as well as surface imagery')
        )
        return proj



"""client.py contains the functionality to link the python steno3d
client with the steno3d website
"""

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals

from builtins import input
from builtins import str
from functools import wraps
from os import mkdir
from os import path
from time import sleep

import requests
from six import string_types
from six.moves.urllib.parse import urlparse

from .user import User


__version__ = '0.2.14b0'

PRODUCTION_BASE_URL = 'https://steno3d.com/'
SLEEP_TIME = .75

DEVKEY_PROMPT = "If you have a Steno3D developer key, please enter it here > "

WELCOME_MESSAGE = """
>> Welcome to the Python client library for Steno3D!
"""

FIRST_LOGIN = """
If you do not have a Steno3D developer key, you need to request
one from the Steno3D website in order to access the API. Please
log in to the application (if necessary) and request a new key.

{base_url}settings/developer

If you are not yet signed up, you can do that here:

{base_url}signup

When you are ready, please enter the key above, or reproduce this
prompt by calling steno3d.login().

"""

LOGIN_FAILED = """
>> Oh no! We could not log you in. The API developer key that you provided
could not be validated. If your current API key has been lost or
invalidated, please request a new one at:

{base_url}settings/developer

Then, try `steno3d.login('YOUR-NEW-DEVEL-KEY')`

If the problem persists:

1) Restart your Python kernel and try again
2) Update steno3d with `pip install --upgrade steno3d`
3) Ask for <help@steno3d.com>
4) Open an issue https://github.com/3ptscience/steno3dpy/issues

"""

NOT_CONNECTED = """
>> Oh no! We could not connect to the Steno3D server. Please ensure that you are:

1) Connected to the Internet
2) Can connect to Steno3D at https://steno3d.com
3) If you are getting an InsecurePlatformWarning while using pip try:
    a) Upgrading to Python 2.7.9 or above
    b) Or `pip install --upgrade requests[security]`
4) Ask for <help@steno3d.com>
5) Open an issue https://github.com/3ptscience/steno3dpy/issues

"""

BAD_API_KEY = """
>> Oh no! Your API developer key format is incorrect.

It should be your username followed by '//' then 36 characters.
You may also use only your username if you have access to local saved
credentials. If you have not requested an API key or if you have lost
your API key, please request a new one at:

{base_url}settings/developer

"""

INVALID_VERSION = """
Oh no! Your version of steno3d is out of date.

{your_version}
{current_version}

Please update steno3d with `pip install --upgrade steno3d`.

"""

BETA_TEST = """
It looks like you are using a beta version of steno3d. Thank you for
trying it out!

Please, if you run into any problems or have any feedback, open an
issue at https://github.com/3ptscience/steno3dpy/issues

If you would like to switch to the most recent stable version:
> pip uninstall steno3d
> pip install steno3d

"""

ALREADY_LOGGED_IN = """
You are already logged in as @{user}. To log in as a different user
please `steno3d.logout()`, then login specifying a different
username or API developer key.

"""


def needs_login(func):
    """Wrapper used around functions that need you to be logged in"""
    @wraps(func)
    def func_wrapper(*args, **kwargs):
        if not Comms.user.logged_in:
            print("Please login: 'steno3d.login()'")
        else:
            return func(*args, **kwargs)
    return func_wrapper


def pause():
    """Brief pause on localhost to simulate network delay"""
    if 'localhost' in Comms.base_url:
        sleep(SLEEP_TIME)


class _Comms(object):
    """Comms controls the interaction between the python client and the
    Steno3D website.
    """

    def __init__(self):
        self.user = User()
        self._base_url = PRODUCTION_BASE_URL
        self._hard_devel_key = None

    @property
    def host(self):
        """hostname of url"""
        parseresult = urlparse(self.base_url)
        return parseresult.hostname

    @property
    def base_url(self):
        """base url endpoint for uploading"""
        return getattr(self, '_base_url', PRODUCTION_BASE_URL)

    @base_url.setter
    def base_url(self, value):
        assert isinstance(value, string_types), \
            'Endpoint path must be a string'
        # Patch '/' onto bare URL endpoints
        if not value[-1] == '/':
            value += '/'
        # Check for HTTPS
        parsed = urlparse(value)
        if '.com' in parsed.hostname and parsed.scheme != 'https':
            raise Exception('Live endpoints require HTTPS.')

        self._base_url = value

    def login(self, devel_key=None, credentials_file=None,
              skip_credentials=False, endpoint=None):
        """Login to steno3d.com to allow uploading resources. To obtain an
        API developer key, you need a Steno3D account:

        https://steno3d.com/signup

        Then, you can request a devel key:

        https://steno3d.com/settings/developer

        Unless you choose to 'skip_credentials', your API key will be
        saved locally and read next time you call `steno3d.login()`.
        You can always login using a different devel key (or username if
        the corresponding devel key is saved in the credentials file).

        Optional arguments:
            devel_key        - API key from steno3d.com. Prompt will appear if
                               this is not provided or saved in credential
                               file. This may also be a username corresponding
                               to a devel key saved in the credentials file
            credentials_file - Local file where devel keys are stored.
                               (Default: ~/.steno3d_client/credentials)
            skip_credentials - If False (default), devel key will be read
                               from and written to local credentials file.
                               If True, only uses the provided devel key or
                               prompts for a new key.
            endpoint         - Target site (Default: steno3d.com)
        """

        # Check user
        if self.user.logged_in:
            print(ALREADY_LOGGED_IN.format(user=self.user.username))
            return

        # Set endpoint
        if endpoint is not None:
            self.base_url = str(endpoint)

        # Check client version
        if not self._version_ok():
            print('Login failed.')
            return

        print(WELCOME_MESSAGE)

        # Assess credential file options.
        if skip_credentials:
            self._login_with(devel_key)
            return

        # Extract credential file
        if credentials_file is None:
            credentials_file = path.sep.join([path.expanduser('~'),
                                              '.steno3d_client',
                                              'credentials'])
            cred_dir = path.sep.join(credentials_file.split(path.sep)[:-1])
            if not path.isdir(cred_dir):
                mkdir(cred_dir)
        elif isinstance(credentials_file, string_types):
            credentials_file = path.realpath(path.expanduser(
                credentials_file
            ))
            cred_dir = path.sep.join(credentials_file.split(path.sep)[:-1])
        else:
            raise ValueError('credentials_file: must be the name of a file')

        if not path.isdir(cred_dir):
            raise ValueError(
                '{}: credentials file directory must exist'.format(cred_dir)
            )
        if path.exists(credentials_file) and not path.isfile(credentials_file):
            raise ValueError(
                '{}: credentials file must be a file'.format(credentials_file)
            )
        if path.isfile(credentials_file):
            print('Credentials file found: {}'.format(credentials_file))
            with open(credentials_file, 'r') as cred:
                devel_keys = cred.readlines()
            devel_keys = [dk.strip() for dk in devel_keys
                          if self.is_key(dk.strip())]
            usernames = [dk.split('//')[0] for dk in devel_keys]
        else:
            print('Creating new credentials file: {}'.format(credentials_file))
            devel_keys = []
            usernames = []

        # Get key from credential file
        if devel_key in usernames:
            print('Accessing API developer key for @{}'.format(devel_key))
            devel_key = devel_keys[usernames.index(devel_key)]

        if devel_key is None and len(devel_keys) > 0:
            print('Accessing API developer key for @{}'.format(usernames[0]))
            devel_key = devel_keys[0]

        self._login_with(devel_key)

        # Update credential file
        if self.user.logged_in:
            updated_devel_keys = [self.user.devel_key]
            for i, key in enumerate(devel_keys):
                if (key == self.user.devel_key or
                        usernames[i] == self.user.username):
                    continue
                updated_devel_keys += [key]
            with open(credentials_file, 'w') as cred:
                cred.writelines(['{}\n'.format(k) for k in updated_devel_keys])

    def _version_ok(self):
        """Check current Steno3D client version in the database"""
        try:
            resp = requests.post(
                self.base_url + 'api/client/steno3dpy',
                dict(version=__version__)
            )
        except requests.ConnectionError:
            print(NOT_CONNECTED)
            return False
        if resp.status_code == 200:
            resp_json = resp.json()
            your_ver_str = resp_json['your_version']
            your_ver = [int(v) for v in your_ver_str.split('.')]
            curr_ver_str = resp_json['current_version']
            curr_ver = [int(v) for v in curr_ver_str.split('.')]
            if resp_json['valid'] or curr_ver_str == '0.0.0':
                return True
            elif your_ver[0] == curr_ver[0] and your_ver[1] == curr_ver[1]:
                print(INVALID_VERSION.format(
                    your_version='Your version: ' + your_ver_str,
                    current_version='Current version: ' + curr_ver_str
                ))
                return True
            else:
                print(INVALID_VERSION.format(
                    your_version='Your version: ' + your_ver_str,
                    current_version='Required version: ' + curr_ver_str
                ))
                if your_ver[0] < curr_ver[0]:
                    return False
                if your_ver[0] == curr_ver[0] and your_ver[1] < curr_ver[1]:
                    return False
                return True
        elif resp.status_code == 400:
            resp_json = resp.json()
            if 'b' in __version__.split('.')[2]:
                print(BETA_TEST)
            else:
                print(INVALID_VERSION.format(
                    your_version='Your version: ' + __version__,
                    current_version='Error: ' + resp_json['reason']
                ))
        return True

    def _login_with(self, devel_key):
        """Login with devel_key"""
        if devel_key is None:
            print(FIRST_LOGIN.format(base_url=self.base_url))
            try:
                devel_key = raw_input(DEVKEY_PROMPT)
            except NameError:
                devel_key = input(DEVKEY_PROMPT)
        if not self.is_key(devel_key):
            print(BAD_API_KEY.format(base_url=self.base_url))
            return
        try:
            resp = requests.get(
                self.base_url + 'api/me',
                headers={'sshKey': devel_key,
                         'client': 'steno3dpy:{}'.format(__version__)}
            )
        except requests.ConnectionError:
            print(NOT_CONNECTED)
            return
        if resp.status_code is not 200:
            print(LOGIN_FAILED.format(base_url=self.base_url))
            self.logout()
            return
        self.user.login_with_json(resp.json())
        self.user.set_key(devel_key)
        self._cookies = dict(resp.cookies)
        print(
            'Welcome to Steno3D! You are logged in as @{name}'.format(
                name=self.user.username
            )
        )

    @staticmethod
    def is_key(devel_key):
        """Checks if devel_key is a valid API key string"""
        if not isinstance(devel_key, string_types):
            return False
        split_key = devel_key.split('//')
        return len(split_key) == 2 and len(split_key[1]) == 36

    def logout(self):
        """Logout current user"""
        if self.user.logged_in:
            print('Logging out of steno3d...')
            _Comms.get('signout')
            print('Goodbye, @{}.'.format(self.user.username))
        self._base_url = PRODUCTION_BASE_URL
        self.user.logout()

    @staticmethod
    def post(url, data=None, files=None):
        """Post data and files to the steno3d online endpoint"""
        return _Comms._communicate(requests.post, url, data, files)

    @staticmethod
    def put(url, data=None, files=None):
        """Put data and files to the steno3d online endpoint"""
        return _Comms._communicate(requests.put, url, data, files)

    @staticmethod
    def get(url):
        """Make a get request from a steno3d online endpoint"""
        return _Comms._communicate(requests.get, url, None, None)

    @staticmethod
    def _communicate(request_fcn, url, data, files):
        """Post data and files to the steno3d online endpoint"""
        data = {} if data is None else data
        files = {} if files is None else files
        filedict = {}
        for filename in files:
            if hasattr(files[filename], 'dtype'):
                filedict[filename] = files[filename].file
                filedict[filename + 'Type'] = files[filename].dtype
            else:
                filedict[filename] = files[filename]
        req = request_fcn(
            Comms.base_url + url,
            data=data,
            files=filedict,
            headers={'sshKey': Comms.user.devel_key,
                     'client': 'steno3dpy:{}'.format(__version__)},
            cookies=Comms._cookies
        )
        if req.status_code < 210:
            Comms._cookies.update(req.cookies)
        for key in files:
            files[key].file.close()

        try:
            resp = req.json()
        except ValueError:
            resp = req
        return {"status_code": req.status_code, "json": resp}


Comms = _Comms()


def plot(url):
    """Return an IFrame plot"""
    from IPython.display import IFrame
    return IFrame(url, width='100%', height=500)



"""There are parser modules available for Steno3D to read certain file
types. When a parser module is imported, the available parsers appear
inside `steno3d.parsers`.
"""

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals

from os.path import expanduser as _expanduser
from os.path import isfile as _isfile
from os.path import realpath as _realpath

from six import with_metaclass as _with_metaclass
from six import string_types as _string_types

from .project import Project as _Project
from .traits import HasSteno3DTraits as _HasSteno3DTraits
from .traits import String as _String
from .traits import KeywordInstance as _KWInst


class _ParserMetaClass(_HasSteno3DTraits.__class__):
    """metaclass ParserMetaclass

    Metaclass to ensure Parser classes fit the required format and
    get added to the steno3d.parsers namespace
    """

    def __new__(mcs, name, bases, attrs):
        if name != 'BaseParser':
            assert 'extensions' in attrs, \
                "You must give the `extensions` that are supported by the " \
                "{name} parser, e.g. ('obj',)".format(name=name)
            exts = attrs['extensions']
            assert isinstance(exts, tuple) and len(exts) > 0, \
                'The `extensions` in the {name} parser must be a tuple of ' \
                'supported extensions'.format(name=name)
            for ext in exts:
                assert isinstance(ext, _string_types), \
                    'Extensions in the {name} parser must be ' \
                    'strings'.format(name=name)
            assert '__init__' not in attrs, \
                "The BaseParser __init__ function takes the file_name and " \
                "any additional keyword input. Please perform other " \
                "initialization tasks for the {name} parser in " \
                "_initialize()".format(name=name)
            assert 'parse' in attrs and callable(attrs['parse']), \
                "Parser class {name} must contain a parse() method"
        new_class = super(_ParserMetaClass, mcs).__new__(
            mcs, name, bases, attrs
        )
        globals()[name] = new_class
        return new_class


class BaseParser(_with_metaclass(_ParserMetaClass,
                                 _HasSteno3DTraits)):
    """Base class for Steno3D parser objects

    BaseParser itself cannot be instantiated. Please use the specific
    parser corresponding to your file type or `AllParsers` instead.
    For more information about parsers, see
    https://python.steno3d.com/en/latest/content/parser.html
    """

    extensions = (None,)
    file_name = _String(
        help='The file to parse'
    )
    project = _KWInst(
        help='The project to parse file_name into',
        klass=_Project,
        kw=dict(description='Project imported from parsed file')
    )

    def __init__(self, file_name):
        super(BaseParser, self).__init__()
        if self.extensions[0] is None:
            raise ParseError(
                '\nPlease use the specific parser corresponding to your '
                'file type or \n`AllParsers`, not `BaseParser`. For more '
                'information about parsers, see\n'
                'https://python.steno3d.com/en/latest/content/parsers.html'
            )
        self.file_name = self._validate_file(file_name)
        self._initialize()

    def _validate_file(self, file_name):
        """function _validate_file

        Input:
            file_name - The file to be validated

        Output:
            validated file_name

        _validate_file verifies the file exists and the extension matches
        the parser extension(s) before proceeding. This hook can be
        overwritten to perform different file checks or remove the checks
        entirely as long as it returns the file_name.
        """
        if not isinstance(file_name, _string_types):
            raise ParseError(
                '{}: file_name must be a string'.format(file_name)
            )
        file_name = _realpath(_expanduser(file_name))
        if not _isfile(file_name):
            raise ParseError('{}: File not found.'.format(file_name))
        fnsplit = file_name.split('.')
        if len(fnsplit) < 2 or fnsplit[-1] not in self.extensions:
            raise ParseError('{name}: Unsupported extension. Supported '
                             'extensions are {exts}'.format(
                                 name=file_name,
                                 exts='(' + ', '.join(self.extensions) + ')'
                             ))
        return file_name

    def _initialize(self):
        """function _initialize

        _initialize is a hook that is called during parser __init__
        after _validate_file. It can be overwritten to perform any
        additional startup tasks
        """

    def parse(self, project=None, **kwargs):
        """function parse

        Parses the file provided at parser instantiation into a
        Steno3D project.

        Optional input:
            project - Preexisting project to add resources to. If not
                      provided, a new project will be created.

        Output:
            tuple of Steno3D project(s) parsed from file_name
        """
        raise NotImplementedError()


class _AllParserMetaClass(_HasSteno3DTraits.__class__):
    """metaclass AllParserMetaClass

    Metaclass to ensure AllParser classes fit the requried format and
    get added to the steno3d.parsers namespace
    """

    def __new__(mcs, name, bases, attrs):
        if name != 'AllParsers':
            assert 'extensions' in attrs, \
                "{name} must contain a dictionary of extensions and " \
                "parsers named `extensions`".format(name=name)
            assert isinstance(attrs['extensions'], dict), \
                "{name} extensions must be a dictionary of extensions and " \
                "supporting parser".format(name=name)
            for ext in attrs['extensions']:
                assert isinstance(ext, _string_types), \
                    'Extensions in {name} must be strings'.format(name=name)
                assert issubclass(type(attrs['extensions'][ext]),
                                  _ParserMetaClass), \
                    'Extensions in {name} must direct to a ' \
                    'Parser class'.format(name=name)
        new_class = super(_AllParserMetaClass, mcs).__new__(
            mcs, name, bases, attrs
        )
        globals()[name] = new_class
        return new_class


class AllParsers(_with_metaclass(_AllParserMetaClass,
                                 _HasSteno3DTraits)):
    """class AllParsers

    Steno3D parser class that selects the appropriate parser for a file
    based on the file's extension. File type must have a corresponding
    parser imported.
    """

    def __new__(cls, file_name, **kwargs):
        if getattr(cls, 'extensions', None) is None:
            cls.extensions = dict()
            parser_keys = [
                k for k in globals()
                if (k != 'BaseParser' and
                    issubclass(type(globals()[k]), _ParserMetaClass))
            ]
            if len(parser_keys) == 0:
                raise ParseError(
                    '\nNo parsers imported! For more information on how to '
                    'use parsers, please see\n'
                    'https://python.steno3d.com/en/latest/content/parsers.html'
                )
            for k in parser_keys:
                for ext in globals()[k].extensions:
                    if ext not in cls.extensions:
                        cls.extensions[ext] = globals()[k]
                    elif issubclass(type(cls.extensions[ext]),
                                    _ParserMetaClass):
                        cls.extensions[ext] = (cls.extensions[ext].__name__ +
                                               ', ' + globals()[k].__name__)
                    else:
                        cls.extensions[ext] = (cls.extensions[ext] +
                                               ', ' + globals()[k].__name__)
        fnsplit = file_name.split('.')
        for ext in cls.extensions:
            if len(fnsplit) > 1 and fnsplit[-1] == ext:
                if issubclass(type(cls.extensions[ext]), _ParserMetaClass):
                    return cls.extensions[ext](file_name, **kwargs)

                raise ParseError(
                    '{ext}: file type supported by more than one parser. '
                    'Please specify one of ({parsers})'.format(
                        ext=ext,
                        parsers=cls.extensions[ext]
                    )
                )

        raise ParseError(
            '{bad}: unsupported file extensions. Must be in ({ok})'.format(
                bad=fnsplit[-1] if len(fnsplit) > 1 else file_name,
                ok=', '.join(list(cls.extensions))
            )
        )


class ParseError(IOError):
    """class ParseError

    Custom exception to raise for errors during file parsing
    """

try:
    del absolute_import, division, print_function, unicode_literals
except NameError:
    # Error cleaning namespace
    pass



from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals

from .airports import Airports
from .brain import Brain
from .misc import Images, Files
from .rocket import Rocket
from .teapot import Teapot
from .topography import Topography
from .tsyganenko import Tsyganenko
from .wolfpass import Wolfpass

try:
    del airports, misc, teapot, topography, wolfpass
    del base
    del absolute_import, division, print_function, unicode_literals
except NameError:
    # Error cleaning namespace
    pass



"""rocket.py provides an example Steno3D project of rocket engine surface"""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals

import numpy as np

from .base import BaseExample, exampleproperty
from ..project import Project
from ..surface import Mesh2D, Surface


class Rocket(BaseExample):
    """Class containing components of rocket engine surface project.

    Construct the project with Rocket.get_project()
    """

    @exampleproperty
    def filenames(self):
        """all rocket engine files"""
        return ['vertices.npz', 'triangles.npz', 'license.txt']

    @exampleproperty
    def vertices(self):
        """rocket engine vertex file"""
        return Rocket.fetch_data(filename='vertices.npz',
                                 download_if_missing=False,
                                 verbose=False)

    @exampleproperty
    def triangles(self):
        """rocket engine triangle  file"""
        return Rocket.fetch_data(filename='triangles.npz',
                                 download_if_missing=False,
                                 verbose=False)

    @classmethod
    def get_project(self):
        """return rocket engine surface project"""
        proj = Project(
            title='F-1 Rocket Engine',
            description='Engine used in the Saturn V. Model by Hlynkacg'
        )
        verts = np.load(Rocket.vertices)
        tris = np.load(Rocket.triangles)
        for arr_name in tris.files:
            Surface(
                project=proj,
                mesh=Mesh2D(
                    vertices=verts[arr_name],
                    triangles=tris[arr_name],
                ),
                title='Rocket Engine Surface: {}'.format(arr_name)
            )
        return proj



"""point.py contains the Point composite resource for steno3d"""

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals

from six import string_types

from numpy import ndarray
from traitlets import observe, validate

from .base import BaseMesh
from .base import CompositeResource
from .data import DataArray
from .options import ColorOptions
from .options import Options
from .texture import Texture2DImage
from .traits import Array, HasSteno3DTraits, KeywordInstance, Repeated, String


class _Mesh0DOptions(Options):
    pass


class _PointOptions(ColorOptions):
    pass


class Mesh0D(BaseMesh):
    """Contains spatial information of a 0D point cloud."""
    vertices = Array(
        help='Point locations',
        shape=('*', 3),
        dtype=float
    )
    opts = KeywordInstance(
        help='Mesh0D Options',
        klass=_Mesh0DOptions,
        allow_none=True
    )

    @property
    def nN(self):
        """ get number of nodes """
        return len(self.vertices)

    def _nbytes(self, arr=None):
        if arr is None or (isinstance(arr, string_types) and
                           arr == 'vertices'):
            arr = self.vertices
        if isinstance(arr, ndarray):
            return arr.astype('f4').nbytes
        raise ValueError('Mesh0D cannot calculate the number of '
                         'bytes of {}'.format(arr))

    @observe('vertices')
    def _reject_large_files(self, change):
        try:
            self._validate_file_size(change['name'], change['new'])
        except ValueError as err:
            setattr(change['owner'], change['name'], change['old'])
            raise err

    @validate('vertices')
    def _validate_verts(self, proposal):
        """Check if mesh content is built correctly"""
        proposal['owner']._validate_file_size('vertices', proposal['value'])
        return proposal['value']

    def _get_dirty_files(self, force=False):
        files = super(Mesh0D, self)._get_dirty_files(force)
        dirty = self._dirty_traits
        if 'vertices' in dirty or force:
            files['vertices'] = \
                self.traits()['vertices'].serialize(self.vertices)
        return files

    @classmethod
    def _build_from_json(cls, json, **kwargs):
        mesh = Mesh0D(
            title=kwargs['title'],
            description=kwargs['description'],
            vertices=Array.download(
                url=json['vertices'],
                shape=(json['verticesSize']//12, 3),
                dtype=json['verticesType']
            ),
            opts=json['meta']
        )
        return mesh

    @classmethod
    def _build_from_omf(cls, omf_geom, omf_project):
        mesh = Mesh0D(
            vertices=(omf_geom.vertices.array +
                      omf_geom.origin +
                      omf_project.origin)
        )
        return mesh


class _PointBinder(HasSteno3DTraits):
    """Contains the data on a 0D point cloud"""
    location = String(
        help='Location of the data on mesh',
        default_value='N',
        choices={
            'N': ('NODE', 'CELLCENTER', 'CC', 'VERTEX')
        }
    )
    data = KeywordInstance(
        help='Data',
        klass=DataArray
    )


class Point(CompositeResource):
    """Contains all the information about a 0D point cloud"""
    mesh = KeywordInstance(
        help='Mesh',
        klass=Mesh0D
    )
    data = Repeated(
        help='Data',
        trait=KeywordInstance(klass=_PointBinder),
        allow_none=True
    )
    textures = Repeated(
        help='Textures',
        trait=KeywordInstance(klass=Texture2DImage),
        allow_none=True
    )
    opts = KeywordInstance(
        help='Options',
        klass=_PointOptions,
        allow_none=True
    )

    def _nbytes(self):
        return (self.mesh._nbytes() +
                sum(d.data._nbytes() for d in self.data) +
                sum(t._nbytes() for t in self.textures))

    @validate('data')
    def _validate_data(self, proposal):
        """Check if resource is built correctly"""
        for ii, dat in enumerate(proposal['value']):
            assert dat.location == 'N'
            valid_length = proposal['owner'].mesh.nN
            if len(dat.data.array) != valid_length:
                raise ValueError(
                    'point.data[{index}] length {datalen} does not match '
                    '{loc} length {meshlen}'.format(
                        index=ii,
                        datalen=len(dat.data.array),
                        loc=dat.location,
                        meshlen=valid_length
                    )
                )
        return proposal['value']


__all__ = ['Point', 'Mesh0D']



"""texture.py contains the texture resource structures"""

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals

from collections import namedtuple
from io import BytesIO
from json import dumps
from six import string_types
from traitlets import observe, validate

from .base import BaseTexture2D
from .traits import Image, Vector


FileProp = namedtuple('FileProp', ['file', 'dtype'])


class Texture2DImage(BaseTexture2D):
    """Contains an image that can be mapped to a 2D surface"""

    _resource_class = 'image'

    O = Vector(
        help='Origin of the texture'
    )
    U = Vector(
        help='U axis of the texture'
    )
    V = Vector(
        help='V axis of the texture'
    )
    image = Image(
        help='Image file'
    )

    def _nbytes(self, img=None):
        if img is None or (isinstance(img, string_types) and img == 'image'):
            img = self.image
        try:
            img.seek(0)
            return len(img.read())
        except:
            raise ValueError('Texture2DImage cannot calculate the number of '
                             'bytes of {}'.format(img))

    @observe('image')
    def _reject_large_files(self, change):
        try:
            self._validate_file_size(change['name'], change['new'])
        except ValueError as err:
            setattr(change['owner'], change['name'], change['old'])
            raise err

    @validate('image')
    def _validate_Z(self, proposal):
        proposal['owner']._validate_file_size('image', proposal['value'])
        return proposal['value']

    def _get_dirty_files(self, force=False):
        files = super(Texture2DImage, self)._get_dirty_files(force)
        dirty = self._dirty_traits
        if 'image' in dirty or force:
            self.image.seek(0)
            copy = BytesIO()
            copy.name = 'texture_copy.png'
            copy.write(self.image.read())
            copy.seek(0)
            files['image'] = FileProp(copy, 'png')
        return files

    def _get_dirty_data(self, force=False):
        datadict = super(Texture2DImage, self)._get_dirty_data(force)
        dirty = self._dirty_traits
        if ('O' in dirty or 'U' in dirty or 'V' in dirty) or force:
            datadict['OUV'] = dumps(dict(
                O=self.O.tolist(),
                U=self.U.tolist(),
                V=self.V.tolist(),
            ))
        return datadict

    def _repr_png_(self):
        """For IPython display"""
        if self.image is None:
            return None
        self.image.seek(0)
        return self.image.read()

    @classmethod
    def _build_from_json(cls, json, **kwargs):
        tex = Texture2DImage(
            title=kwargs['title'],
            description=kwargs['description'],
            O=json['OUV']['O'],
            U=json['OUV']['U'],
            V=json['OUV']['V'],
            image=Image.download(json['image'])
        )
        return tex

    @classmethod
    def _build_from_omf(cls, omf_tex, omf_project):
        tex = Texture2DImage(
            title=omf_tex.name,
            description=omf_tex.description,
            origin=omf_tex.origin + omf_project.origin,
            U=omf_tex.axis_u,
            V=omf_tex.axis_v,
            image=omf_tex.image
        )
        return tex


__all__ = ['Texture2DImage']
